#ifndef _IRI_JOINTS_MODULE_BT_H
#define _IRI_JOINTS_MODULE_BT_H

#include <iri_base_algorithm/iri_base_algorithm.h>
// #include <behaviortree_cpp_v3/bt_factory.h>
// #include <behaviortree_cpp_v3/behavior_tree.h>
// #include "behaviortree_cpp_v3/blackboard.h"

#include "iri_bt_factory.h"
#include <iri_joints_module/iri_joints_module.h>

// template <> inline std::vector<std::string> convertFromString(BT::StringView str)
// {
//   auto parts = BT::splitString(str, ',');
//   std::vector<std::string> joint_names;
//   for (uint8_t i=0; i < parts.size(); i++)
//     joint_names.push_back(parts[i]);
//   return joint_names;
// }

// template <> inline std::vector<double> convertFromString(BT::StringView str)
// {
//   auto parts = BT::splitString(str, ',');
//   std::vector<double> d;
//   for (uint8_t i=0; i < parts.size(); i++)
//     d.push_back(BT::convertFromString<double>(parts[i]));
//   return d;
// }

class CIriJointsModuleBT
{
  private:
    // object of iri_joints_module
    CIriJointsModule joints_module;

  public:
    /**
      * \brief Constructor
      *
      */
    CIriJointsModuleBT();

    /**
      * \brief Register all conditions and actions needed for using the module
      *
      * This function registers all conditions and actions needed for using the module
      * to the factory provided including all input and output ports required
      * by actions.
      *
      * \param factory (IriBehaviorTreeFactory)
      *
      */
    void init(IriBehaviorTreeFactory &factory);

    /**
    * \brief Destructor
    *
    * This destructor frees all necessary dynamic memory allocated within this
    * this class.
    */
    ~CIriJointsModuleBT();

  protected:

    ////Point_to Action
    /**
      * \brief Synchronized point_to function.
      *
      * This function calls point_to of iri_joints_module. As this is a
      * synchronous action is_point_to_finished() must be used to know when the action
      * has actually finished.
      *
      * It has the following input ports:
      *
      *   x (double): Point X coordenate on its reference frame.
      *
      *   y (double): Point Y coordenate on its reference frame.
      *
      *   z (double): Point Z coordenate on its reference frame.
      *
      *   max_vel (double): The maximum velocity for the servos.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus sync_point_to(BT::TreeNode& self);

    /**
      * \brief Async point_to function.
      *
      * This function calls point_to of iri_joints_module. As this is
      * an asynchronous action is_point_to_finished() is checked to know when the action
      * has actually finished.
      *
      * It has the following input ports:
      *
      *   x (double): Point X coordenate on its reference frame.
      *
      *   y (double): Point Y coordenate on its reference frame.
      *
      *   z (double): Point Z coordenate on its reference frame.
      *
      *   max_vel (double): The maximum velocity for the servos.
      *
      * It also has a bidirectional port:
      *
      *   new_goal (bool): If it's a new_goal.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull and the action has finished (BT:NodeStatus::SUCCESS) or the
      * request has been unsuccessfull (BT:NodeStatus::FAILURE) or it is still
      * in progress (BT:NodeStatus::RUNNING). If inputs are missing or incorrect
      * it also returns BT:NodeStatus::FAILURE
      *
      */
    BT::NodeStatus async_point_to(BT::TreeNode& self);

    /**
      * \brief Synchronized cancel_point_to function.
      *
      * This function calls cancel_point_to of iri_joints_module and waits to be finished.
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus cancel_point_to(void);

    /**
      * \brief Checks if the last point_to goal has finished or not.
      *
      * This function must be used when sync_point_to is called.
      *
      * \return a BT:NodeStatus indicating whether the last point_to goal has
      * ended (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE),
      * regardless of its success or not.
      *
      */
    BT::NodeStatus is_point_to_finished(void);

    /**
      * \brief Checks if the last point_to goal has finished or not.
      *
      * This function must be used when sync_point_to is called.
      *
      * \return a BT:NodeStatus indicating whether the last point_to goal has
      * ended (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::RUNNING),
      * regardless of its success or not.
      *
      */
    BT::NodeStatus async_is_point_to_finished(void);


    ///Move to angles action
    /**
      * \brief Synchronized move_to angles function.
      *
      * This function calls move_to angles of iri_joints_module. As this is a
      * synchronous action is_move_to_angles_finished() must be used to know when the action
      * has actually finished.
      *
      * It has the following input ports:
      *
      *   joint_names (std::vector<std::string>): Name of each joint desired to move.
      *
      *   angles (std::vector<double>): Derired angle of each joint.
      *
      *   vel (std::vector<double>): Derired velocity of each joint.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus sync_move_to_angles(BT::TreeNode& self);

    /**
      * \brief Async move_to angles function.
      *
      * This function calls move_to angles of iri_joints_module. As this is
      * an asynchronous action is_move_to_angles_finished() is checked to know when the action
      * has actually finished.
      *
      * It has the following input ports:
      *
      *   joint_names (std::vector<std::string>): Name of each joint desired to move.
      *
      *   angles (std::vector<double>): Derired angle of each joint.
      *
      *   vel (std::vector<double>): Derired velocity of each joint.
      *
      * It also has a bidirectional port:
      *
      *   new_goal (bool): If it's a new_goal.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull and the action has finished (BT:NodeStatus::SUCCESS) or the
      * request has been unsuccessfull (BT:NodeStatus::FAILURE) or it is still
      * in progress (BT:NodeStatus::RUNNING). If inputs are missing or incorrect
      * it also returns BT:NodeStatus::FAILURE
      *
      */
    BT::NodeStatus async_move_to_angles(BT::TreeNode& self);

    /**
      * \brief Synchronized cancel move_to_angles function.
      *
      * This function calls cancel_move_to_angles of iri_joints_module and waits to be finished.
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus cancel_move_to_angles(void);

    /**
      * \brief Checks if the last move_to angles goal has finished or not.
      *
      * This function must be used when sync_point_to is called.
      *
      * \return a BT:NodeStatus indicating whether the last move_to angles goal has
      * ended (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE),
      * regardless of its success or not.
      *
      */
    BT::NodeStatus is_move_to_angles_finished(void);

    /**
      * \brief Checks if the last move_to angles goal has finished or not.
      *
      * This function must be used when sync_point_to is called.
      *
      * \return a BT:NodeStatus indicating whether the last move_to angles goal has
      * ended (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::RUNNING),
      * regardless of its success or not.
      *
      */
    BT::NodeStatus async_is_move_to_angles_finished(void);


    ///PointHeadTracker
    /**
      * \brief Synchronized target update function function.
      *
      * This function calls target_update of iri_joints_module.
      *
      * It has the following input ports:
      *
      *   x (double): Point X coordenate on its reference frame.
      *
      *   y (double): Point Y coordenate on its reference frame.
      *
      *   z (double): Point Z coordenate on its reference frame.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus target_update(BT::TreeNode& self);

    /**
      * \brief Synchronized enable tracker function.
      *
      * This function calls enable_tracker of iri_joints_module and waits to be finished.
      * As this is a synchronous action is_tracker_finished() must be used to know
      * when the action has actually finished.
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus enable_tracker(void);

    /**
      * \brief Synchronized cancel tracker function.
      *
      * This function calls cancel_tracker of iri_joints_module and waits to be finished.
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus cancel_tracker(void);

    /**
      * \brief Checks if the tracker has finished or not.
      *
      * This function must be used after enabling the tracker.
      *
      * \return a BT:NodeStatus indicating whether the tracker has
      * finished (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE),
      * regardless of its success or not.
      *
      */
    BT::NodeStatus is_tracker_finished(void);

    /**
      * \brief Checks if the tracker has finished or not.
      *
      * This function must be used after enabling the tracker.
      *
      * \return a BT:NodeStatus indicating whether the tracker has
      * finished (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::RUNNING),
      * regardless of its success or not.
      *
      */
    BT::NodeStatus async_is_tracker_finished(void);


    ///Common
    /**
      * \brief Synchronized set_tolerance function.
      *
      * This function calls set_tolerance of iri_joints_module and waits to be finished.
      *
      * It has the following input ports:
      *
      *   tol (double): Inverse kinematic solver tolerance.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus set_tolerance(BT::TreeNode& self);

    /**
      * \brief Synchronized set_base_frame function.
      *
      * This function calls set_base_frame of iri_joints_module and waits to be finished.
      *
      * It has the following input ports:
      *
      *   base_frame (string): Servo kinematic chain base frame.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus set_base_frame(BT::TreeNode& self);

    /**
      * \brief Synchronized set_urdf_param function.
      *
      * This function calls set_urdf_param of iri_joints_module and waits to be finished.
      *
      * It has the following input ports:
      *
      *   urdf_param (string): Ros param where is stored the urdf model.
      *
      * \param self node with the required inputs defined as follows:
      *
      * \return a BT:NodeStatus indicating whether the request has been
      * successfull (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      * If inputs are missing or incorrect it also returns BT:NodeStatus::FAILURE.
      *
      */
    BT::NodeStatus set_urdf_param(BT::TreeNode& self);

    ///Joints module status
    /**
      * \brief Checks if the module is active and operating properly.
      *
      * This function must be used to know whether the module is active and
      * operating properly, or not.
      *
      * \return a BT:NodeStatus indicating whether the module is active and
      * operating properly (BT:NodeStatus::SUCCESS) or not (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_running(void);

    /**
      * \brief Checks if the last action goal has been reached successfully.
      *
      * This function must be used to know whether the last action goal has
      * been reached successfully or not, once it has finished.
      *
      * \return a BT:NodeStatus indicating whether the last action goal has
      * been reached successfully (BT:NodeStatus::SUCCESS) or not
      * (BT:NodeStatus::FAILURE), regardless of how it has failed.
      *
      */
    BT::NodeStatus is_joints_module_success(void);

    /**
      * \brief Checks if the goal could not be started.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * goal could not be started.
      *
      * \return a BT:NodeStatus indicating whether the goal could not be started
      * (BT:NodeStatus::SUCCESS) or if it failed due to another reason
      * (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_action_start_fail(void);

    /**
      * \brief Checks if the goal did not complete in the allowed time.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * goal did not complete in the allowed time.
      *
      * \return a BT:NodeStatus indicating whether the goal did not complete in
      * the allowed time (BT:NodeStatus::SUCCESS) or if it failed due to another
      * reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_timeout(void);

    /**
      * \brief Checks if the feedback has not been received for a long time.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * feedback has not been received for a long time.
      *
      * \return a BT:NodeStatus indicating whether the feedback has not been
      * received for a long time (BT:NodeStatus::SUCCESS) or if it failed due
      * to another reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_fb_watchdog(void);

    /**
      * \brief Checks if the navigation goal could not be reached.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * navigation goal could not be reached.
      *
      * \return a BT:NodeStatus indicating whether the navigation goal could not
      * be reached (BT:NodeStatus::SUCCESS) or if it failed due to another
      * reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_aborted(void);

    /**
      * \brief Checks if the current navigation goal has been cancelled by
      * another goal.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * current navigation goal has been cancelled by another goal.
      *
      * \return a BT:NodeStatus indicating whether the current navigation goal
      * has been cancelled by another goal (BT:NodeStatus::SUCCESS) or if it
      * failed due to another reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_preempted(void);

    /**
      * \brief Checks if the goal information is not valid.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * goal information is not valid and the goal will not be executed.
      *
      * \return a BT:NodeStatus indicating whether the goal information is not
      * valid (BT:NodeStatus::SUCCESS) or if it failed due to another reason
      * (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_rejected(void);

    /**
      * \brief Checks if it was impossible to set the value of a parameter.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because it was
      * impossible to set the value of a parameter.
      *
      * \return a BT:NodeStatus indicating whether it was impossible to set the
      * value of a parameter (BT:NodeStatus::SUCCESS) or if it failed due to
      * another reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_set_param_fail(void);

    /**
      * \brief Checks if the parameter to be changed is not present in the
      * remote node.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because the
      * parameter to be changed is not present in the remote node.
      *
      * \return a BT:NodeStatus indicating whether the parameter to be changed
      * is not present in the remote node (BT:NodeStatus::SUCCESS) or if it
      * failed due to another reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_param_not_present(void);

    /**
      * \brief Checks if no joint_states information is being received.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because no
      * joint_states information is being received, so it is not possible to get
      * the current position of the robot.
      *
      * \return a BT:NodeStatus indicating whether no joint_states information is
      * being received (BT:NodeStatus::SUCCESS) or if it failed due to another
      * reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_no_joint_states(void);

    /**
      * \brief Checks if no target information is being received in the allowed time.
      *
      * This function must be used to know the type of failure, once the last
      * goal has finished. Specifically, it checks if it failed because no target
      * information has been received in the allowed time.
      *
      * \return a BT:NodeStatus indicating whether no target information has been received in
      * the allowed time (BT:NodeStatus::SUCCESS) or if it failed due to another
      * reason (BT:NodeStatus::FAILURE).
      *
      */
    BT::NodeStatus is_joints_module_target_lost(void);

  };

  #endif
