#include <iri_joints_module/iri_joints_module.h> 
#include <tf/transform_datatypes.h> 

CIriJointsModule::CIriJointsModule(const std::string &name,const std::string &name_space) : CModule(name,name_space),
  point_to_action("point_to",this->module_nh.getNamespace()),
  point_to_reconf("point_to_reconf",this->module_nh.getNamespace()),
  track_action("track", this->module_nh.getNamespace()),
  move_to_angle_action("move_to_angle", this->module_nh.getNamespace())
{
  this->start_operation();

  this->tracker_point_to_publisher_ = this->module_nh.advertise<control_msgs::PointHeadActionGoal>("tracker_point_to", 1);

  //Variables
  this->state = IRI_JOINTS_MODULE_IDLE;
  this->status = IRI_JOINTS_MODULE_SUCCESS;
  this->new_angles = false;
  this->new_point_goal = false;
  this->cancel_tracker_pending = false;
  this->cancel_point_to_pending = false;
  this->cancel_move_to_angles_pending = false;
  this->enable_tracker_pending = false;

  // initialize JointStates subscriber
  this->joint_states_watchdog.reset(ros::Duration(this->config.joint_states_watchdog_time_s));
  this->joint_states_subscriber = this->module_nh.subscribe("joint_states",1,&CIriJointsModule::joint_states_callback,this);
}

/* State machine */
void CIriJointsModule::state_machine(void)
{
  switch (this->state)
  {
    case IRI_JOINTS_MODULE_IDLE:
      if (this->enable_tracker_pending)
      {
        switch (this->track_action.make_request(this->track_goal))
        {
          case ACT_SRV_SUCCESS: 
            this->state = IRI_JOINTS_MODULE_TRACKING;
            this->enable_tracker_pending = false;
            ROS_INFO("CIriJointsModule::state_machine-> PointHeadTrackerAction goal start successfull");
          break;
          case ACT_SRV_PENDING: 
            this->state = IRI_JOINTS_MODULE_IDLE;
            ROS_WARN("CIriJointsModule::state_machine-> PointHeadTrackerAction goal start pending");
          break;
          case ACT_SRV_FAIL: 
            this->state = IRI_JOINTS_MODULE_IDLE;
            this->status = IRI_JOINTS_MODULE_ACTION_START_FAIL;
            this->enable_tracker_pending = false;
            ROS_ERROR("CIriJointsModule::state_machine-> PointHeadTrackerAction goal start failed");
          break;
        }
        break;
      }
      if (this->new_point_goal)
      {
        switch (this->point_to_action.make_request(this->point_to_goal))
        {
          case ACT_SRV_SUCCESS: 
            this->state = IRI_JOINTS_MODULE_POINT_WAIT;
            this->new_point_goal = false;
            ROS_INFO("CIriJointsModule::state_machine-> PointHeadAction goal start successfull");
          break;
          case ACT_SRV_PENDING: 
            this->state = IRI_JOINTS_MODULE_IDLE;
            ROS_WARN("CIriJointsModule::state_machine-> PointHeadAction goal start pending");
          break;
          case ACT_SRV_FAIL: 
            this->state = IRI_JOINTS_MODULE_IDLE;
            this->status = IRI_JOINTS_MODULE_ACTION_START_FAIL;
            this->new_point_goal = false;
            ROS_ERROR("CIriJointsModule::state_machine-> PointHeadAction goal start failed");
          break;
        }
        break;
      }
      if (this->new_angles)
      {
        switch (this->move_to_angle_action.make_request(this->move_to_angle_goal))
        {
          case ACT_SRV_SUCCESS: 
            this->state = IRI_JOINTS_MODULE_MOVE_ANGLES;
            this->new_angles = false;
            ROS_INFO("CIriJointsModule::state_machine-> MoveToAngles goal start successfull");
          break;
          case ACT_SRV_PENDING: 
            this->state = IRI_JOINTS_MODULE_IDLE;
            ROS_WARN("CIriJointsModule::state_machine-> MoveToAngles goal start pending");
          break;
          case ACT_SRV_FAIL: 
            this->state = IRI_JOINTS_MODULE_IDLE;
            this->status = IRI_JOINTS_MODULE_ACTION_START_FAIL;
            this->new_angles = false;
            ROS_ERROR("CIriJointsModule :state_machine-> MoveToAngles goal start failed");
          break;
        }
        break;
      }
    break;
    case IRI_JOINTS_MODULE_POINT_WAIT:
      switch (this->point_to_action.get_state())
      {
        case ACTION_IDLE:
        case ACTION_RUNNING:
          ROS_DEBUG("CIriJointsModule::state_machine-> PointHeadAction running");
          this->status = IRI_JOINTS_MODULE_RUNNING;
        break;
        case ACTION_SUCCESS:
          ROS_INFO("CIriJointsModule::state_machine-> PointHeadAction ended successfully");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_SUCCESS;
          this->point_to_action.stop_timeout();
        break;
        case ACTION_TIMEOUT:
          ROS_ERROR("CIriJointsModule::state_machine-> PointHeadAction did not finish in the allowed time");
          this->point_to_action.cancel();
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_TIMEOUT;
          this->point_to_action.stop_timeout();
        break;
        case ACTION_FB_WATCHDOG:
          ROS_ERROR("CIriJointsModule::state_machine-> No PointHeadAction's feedback received for a long time");
          this->point_to_action.cancel();
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_FB_WATCHDOG;
          this->point_to_action.stop_timeout();
        break;
        case ACTION_ABORTED:
          ROS_ERROR("CIriJointsModule::state_machine-> PointHeadAction failed to complete");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_ABORTED;
          this->point_to_action.stop_timeout();
        break;
        case ACTION_PREEMPTED:
          ROS_WARN("CIriJointsModule::state_machine-> PointHeadAction was interrupted by another request");
          this->state = IRI_JOINTS_MODULE_IDLE;//////necesario?????????
          this->status = IRI_JOINTS_MODULE_PREEMPTED;//////necesario?????????
          this->point_to_action.stop_timeout();////necesario?????
        break;
        case ACTION_REJECTED:
          ROS_ERROR("CIriJointsModule::state_machine-> PointHeadAction was rejected by server");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_REJECTED;
          this->point_to_action.stop_timeout();
        break;
      }
      if (this->new_point_goal)
      {
        this->state = IRI_JOINTS_MODULE_IDLE;
        this->status = IRI_JOINTS_MODULE_PREEMPTED;
        this->point_to_action.stop_timeout();
      }
      else if (this->cancel_point_to_pending)
      {
        this->cancel_point_to_pending = false;
        this->point_to_action.cancel();
      }
    break;
    case IRI_JOINTS_MODULE_TRACKING:
      switch (this->track_action.get_state())
      {
        case ACTION_IDLE:
        case ACTION_RUNNING:
          ROS_DEBUG("CIriJointsModule::state_machine-> PointHeadTrackerAction running");
          this->status = IRI_JOINTS_MODULE_RUNNING;
        break;
        case ACTION_SUCCESS:
          ROS_INFO("CIriJointsModule::state_machine-> PointHeadTrackerAction ended successfully");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_SUCCESS;
          this->track_action.stop_timeout();
        break;
        case ACTION_TIMEOUT:
          ROS_ERROR("CIriJointsModule::state_machine-> PointHeadTrackerAction target info not updated on the allowed time");
          this->track_action.cancel();
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_TARGET_LOST;
          this->track_action.stop_timeout();
        break;
        case ACTION_FB_WATCHDOG:
          ROS_ERROR("CIriJointsModule::state_machine-> No PointHeadTrackerAction's feedback received for a long time");
          this->track_action.cancel();
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_FB_WATCHDOG;
          this->track_action.stop_timeout();
        break;
        case ACTION_ABORTED:
          ROS_ERROR("CIriJointsModule::state_machine-> PointHeadTrackerAction failed to complete");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_ABORTED;
          this->track_action.stop_timeout();
        break;
        case ACTION_PREEMPTED:
          ROS_WARN("CIriJointsModule::state_machine-> PointHeadTrackerAction was interrupted by another request");
          this->state = IRI_JOINTS_MODULE_IDLE;//////necesario?????????
          this->status = IRI_JOINTS_MODULE_PREEMPTED;//////necesario?????????
          this->track_action.stop_timeout();////necesario?????
        break;
        case ACTION_REJECTED:
          ROS_ERROR("CIriJointsModule::state_machine-> PointHeadTrackerAction was rejected by server");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_REJECTED;
          this->track_action.stop_timeout();
        break;
      }
      if (this->new_point_goal)
      {
        this->track_action.update_timeout(this->config.tracker_timeout_s);
        this->tracker_point_to_publisher_.publish(this->target_goal);
        this->new_point_goal = false;
      }
      if (this->enable_tracker_pending)
      {
        this->enable_tracker_pending = false;
        ROS_WARN("CIriJointsModule::state_machine-> PointHeadTrackerAction already enabled.");
      }
      else if (this->cancel_tracker_pending)
      {
        this->cancel_tracker_pending = false;
        this->track_action.cancel();
      }
    break;
    case IRI_JOINTS_MODULE_MOVE_ANGLES:
      switch (this->move_to_angle_action.get_state())
      {
        case ACTION_IDLE:
        case ACTION_RUNNING:
          ROS_DEBUG("CIriJointsModule::state_machine-> MoveToAngles running");
          this->status = IRI_JOINTS_MODULE_RUNNING;
        break;
        case ACTION_SUCCESS:
          ROS_INFO("CIriJointsModule::state_machine-> MoveToAngles ended successfully");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_SUCCESS;
          this->move_to_angle_action.stop_timeout();
        break;
        case ACTION_TIMEOUT:
          ROS_ERROR("CIriJointsModule::state_machine-> MoveToAngles did not finish in the allowed time");
          this->move_to_angle_action.cancel();
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_TIMEOUT;
          this->move_to_angle_action.stop_timeout();
        break;
        case ACTION_FB_WATCHDOG:
          ROS_ERROR("CIriJointsModule::state_machine-> No MoveToAngles's feedback received for a long time");
          this->move_to_angle_action.cancel();
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_FB_WATCHDOG;
          this->move_to_angle_action.stop_timeout();
        break;
        case ACTION_ABORTED:
          ROS_ERROR("CIriJointsModule::state_machine-> MoveToAngles failed to complete");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_ABORTED;
          this->move_to_angle_action.stop_timeout();
        break;
        case ACTION_PREEMPTED:
          ROS_WARN("CIriJointsModule::state_machine-> MoveToAngles was interrupted by another request");
          this->state = IRI_JOINTS_MODULE_IDLE;//////necesario?????????
          this->status = IRI_JOINTS_MODULE_PREEMPTED;//////necesario?????????
          this->move_to_angle_action.stop_timeout();////necesario?????
        break;
        case ACTION_REJECTED:
          ROS_ERROR("CIriJointsModule::state_machine-> MoveToAngles was rejected by server");
          this->state = IRI_JOINTS_MODULE_IDLE;
          this->status = IRI_JOINTS_MODULE_REJECTED;
          this->move_to_angle_action.stop_timeout();
        break;
      }
      if (this->new_angles)
      {
        this->state = IRI_JOINTS_MODULE_IDLE;
        this->status = IRI_JOINTS_MODULE_PREEMPTED;
        this->move_to_angle_action.stop_timeout();
      }
      else if (this->cancel_move_to_angles_pending)
      {
        this->cancel_move_to_angles_pending = false;
        this->move_to_angle_action.cancel();
      }
    break;
  }
  if (this->joint_states_watchdog.is_active())
  {
    this->status = IRI_JOINTS_MODULE_NO_JOINT_STATES;
    ROS_ERROR("CIriJointsModule::state_machine-> No joint_states received.");
  }
}

void CIriJointsModule::reconfigure_callback(iri_joints_module::IriJointsModuleConfig &config, uint32_t level)
{
  ROS_DEBUG("CIriJointsModule::reconfigure_callback-> reconfigure callback");
  this->lock();
  this->config=config;
  /* set the module rate */
  this->set_rate(config.rate_hz);
  /* point_to action parameters */
  this->new_point_goal = false;
  this->point_to_goal.target.header.frame_id = config.point_frame;
  this->target_goal.goal.target.header.frame_id = config.point_frame;
  switch(config.pointing_axis)
  {
    case axis_x:
      this->point_to_goal.pointing_axis.x = 1.0;
      this->point_to_goal.pointing_axis.y = 0.0;
      this->point_to_goal.pointing_axis.z = 0.0;
      this->target_goal.goal.pointing_axis.x = 1.0;
      this->target_goal.goal.pointing_axis.y = 0.0;
      this->target_goal.goal.pointing_axis.z = 0.0;
    break;
    case axis_y:
      this->point_to_goal.pointing_axis.x = 0.0;
      this->point_to_goal.pointing_axis.y = 1.0;
      this->point_to_goal.pointing_axis.z = 0.0;
      this->target_goal.goal.pointing_axis.x = 0.0;
      this->target_goal.goal.pointing_axis.y = 1.0;
      this->target_goal.goal.pointing_axis.z = 0.0;
    break;
    case axis_z:
      this->point_to_goal.pointing_axis.x = 0.0;
      this->point_to_goal.pointing_axis.y = 0.0;
      this->point_to_goal.pointing_axis.z = 1.0;
      this->target_goal.goal.pointing_axis.x = 0.0;
      this->target_goal.goal.pointing_axis.y = 0.0;
      this->target_goal.goal.pointing_axis.z = 1.0;
    break;
  }
  this->point_to_goal.pointing_frame = config.pointing_frame;
  this->target_goal.goal.pointing_frame = config.pointing_frame;

  /*PointTo Action parameters*/
  this->point_to_action.set_max_num_retries(config.pt_max_retries);
  if(config.pt_enable_watchdog)
    this->point_to_action.enable_watchdog(config.pt_feedback_watchdog_time_s);
  else
    this->point_to_action.disable_watchdog();
  if(this->config.pt_enable_timeout)
    this->point_to_action.enable_timeout(config.pt_timeout_s);
  else
    this->point_to_action.disable_timeout();

  if(this->config.pt_enabled)
    this->point_to_action.enable();
  else
    this->point_to_action.disable();


  /*Tracker Action parameters*/
  this->track_action.set_max_num_retries(config.tracker_max_retries);
  if(config.tracker_enable_watchdog)
    this->track_action.enable_watchdog(config.tracker_feedback_watchdog_time_s);
  else
    this->track_action.disable_watchdog();
  if(this->config.tracker_enable_timeout)
    this->track_action.enable_timeout(config.tracker_timeout_s);
  else
    this->track_action.disable_timeout();

  if(this->config.tracker_enabled)
    this->track_action.enable();
  else
    this->track_action.disable();

  /*Move to angles Action parameters*/
  this->move_to_angle_action.set_max_num_retries(config.move_to_angles_max_retries);
  if(config.move_to_angles_enable_watchdog)
    this->move_to_angle_action.enable_watchdog(config.move_to_angles_feedback_watchdog_time_s);
  else
    this->move_to_angle_action.disable_watchdog();
  if(this->config.move_to_angles_enable_timeout)
    this->move_to_angle_action.enable_timeout(config.move_to_angles_timeout_s);
  else
    this->move_to_angle_action.disable_timeout();

  if(this->config.move_to_angles_enabled)
    this->move_to_angle_action.enable();
  else
    this->move_to_angle_action.disable();

  this->unlock();
}

void CIriJointsModule::joint_states_callback(const sensor_msgs::JointState::ConstPtr& msg)
{
  ROS_DEBUG("CIriJointsModule::joint_states_callback-> JointStates callback");

  this->lock();
  // reset the JointStates callback
  this->joint_states_watchdog.reset(ros::Duration(this->config.joint_states_watchdog_time_s));
  // save the current position of the servos
  
  this->unlock();
}

iri_joints_module_status_t CIriJointsModule::get_status(void)
{
  return this->status;
}

/*PointTo action*/
bool CIriJointsModule::point_to(double x, double y, double z, double max_vel)
{
  this->lock();
  if (this->state == IRI_JOINTS_MODULE_TRACKING)
  {
    ROS_WARN("CIriJointsModule::point_to-> Tracking is active. Dropping this goal.");
    this->unlock();
    return false;
  }
  else if (this->state == IRI_JOINTS_MODULE_MOVE_ANGLES)
  {
    ROS_WARN("CIriJointsModule::point_to-> MoveToAngles is active. Dropping this goal.");
    this->unlock();
    return false;
  }
  this->point_to_goal.target.point.x = x;
  this->point_to_goal.target.point.y = y;
  this->point_to_goal.target.point.z = z;
  this->point_to_goal.max_velocity = max_vel;
  this->point_to_goal.target.header.stamp = ros::Time::now();
  this->new_point_goal = true;
  this->unlock();
  return true;
}

void CIriJointsModule::cancel_point_to(void)
{
  if (!point_to_action.is_enabled())
    return;
  if(this->state == IRI_JOINTS_MODULE_POINT_WAIT)
    this->cancel_point_to_pending = true;
  else
    ROS_WARN("CIriJointsModule::cancel_point_to-> PointTo can't be cancelled because is not active.");
}

bool CIriJointsModule::is_point_to_finished(void)
{
  if (!point_to_action.is_enabled())
    return true;
  this->lock();
  if(this->state == IRI_JOINTS_MODULE_POINT_WAIT || (this->state == IRI_JOINTS_MODULE_IDLE && this->new_point_goal))
  {
    this->unlock();
    return false;
  }
  else
  {
    this->unlock();
    return true;
  }
}

/*Tracking action*/
bool CIriJointsModule::target_update(double x, double y, double z)
{
  if (!track_action.is_enabled())
    return true;
  this->lock();
  if (this->state == IRI_JOINTS_MODULE_POINT_WAIT)
  {
    ROS_WARN("CIriJointsModule::target_update-> PointTo action is active. Dropping this goal.");
    this->unlock();
    return false;
  }
  else if (this->state == IRI_JOINTS_MODULE_MOVE_ANGLES)
  {
    ROS_WARN("CIriJointsModule::target_update-> MoveToAngles is active. Dropping this goal.");
    this->unlock();
    return false;
  }
  else if (this->state == IRI_JOINTS_MODULE_IDLE)
  {
    ROS_WARN("CIriJointsModule::target_update-> Tracking is not active. Dropping this goal.");
    this->unlock();
    return false;
  }
  this->target_goal.goal.target.point.x = x;
  this->target_goal.goal.target.point.y = y;
  this->target_goal.goal.target.point.z = z;
  this->target_goal.goal.target.header.stamp = ros::Time::now();
  this->new_point_goal = true;
  this->unlock();
  return true;
}

bool CIriJointsModule::enable_tracker(void)
{
  this->lock();
  if (this->state == IRI_JOINTS_MODULE_POINT_WAIT)
  {
    ROS_ERROR("CIriJointsModule::enable_tracker-> PointTo action is active. Cancel PointTo before start tracking.");
    this->unlock();
    return false;
  }
  else if (this->state == IRI_JOINTS_MODULE_MOVE_ANGLES)
  {
    ROS_ERROR("CIriJointsModule::enable_tracker-> MoveToAngles action is active. Cancel MoveToAngles before start tracking.");
    this->unlock();
    return false;
  }
  else
  {
    this->enable_tracker_pending = true;
    this->unlock();
    return true;
  }
}

void CIriJointsModule::cancel_tracker(void)
{
  if (!track_action.is_enabled())
    return;
  if(this->state == IRI_JOINTS_MODULE_TRACKING)
    this->cancel_tracker_pending = true;
  else
    ROS_WARN("CIriJointsModule::cancel_tracker-> Tracker can't be cancelled because is not active.");
}

bool CIriJointsModule::is_tracker_finished(void)
{
  if (!track_action.is_enabled())
    return true;
  this->lock();
  if(this->state == IRI_JOINTS_MODULE_TRACKING || (this->state == IRI_JOINTS_MODULE_IDLE && this->enable_tracker_pending))
  {
    this->unlock();
    return false;
  }
  else
  {
    this->unlock();
    return true;
  }
}

/*Move to angles action*/
bool CIriJointsModule::move_to_angles(std::vector<std::string> joint_names, std::vector<double> angles, std::vector<double> vel)
{
  this->lock();
  if (this->state == IRI_JOINTS_MODULE_POINT_WAIT)
  {
    ROS_WARN("CIriJointsModule::move_to_angles-> PointTo action is active. Dropping this goal.");
    this->unlock();
    return false;
  }
  else if (this->state == IRI_JOINTS_MODULE_TRACKING)
  {
    ROS_WARN("CIriJointsModule::move_to_angles-> Tracking is active. Dropping this goal.");
    this->unlock();
    return false;
  }
  if (joint_names.size() != angles.size() || joint_names.size() != vel.size())
  {
    ROS_ERROR("CIriJointsModule::move_to_angles-> The size of joint_names, angles and vel must be the same.");
    this->unlock();
    return false;
  }

  this->move_to_angle_goal.trajectory.header.frame_id = this->config.frame_id;
  std::vector<std::string>().swap(this->move_to_angle_goal.trajectory.joint_names);
  this->move_to_angle_goal.trajectory.joint_names = joint_names;

  control_msgs::JointTolerance tol;
  tol.position = this->config.move_to_angles_pos_tol;
  tol.velocity = this->config.move_to_angles_vel_tol;
  tol.acceleration = this->config.move_to_angles_acc_tol;
  std::vector<control_msgs::JointTolerance>().swap(this->move_to_angle_goal.path_tolerance);
  std::vector<control_msgs::JointTolerance>().swap(this->move_to_angle_goal.goal_tolerance);
  for (uint8_t i; i < this->move_to_angle_goal.trajectory.joint_names.size(); i++)
  {
    tol.name = this->move_to_angle_goal.trajectory.joint_names[i];
    this->move_to_angle_goal.path_tolerance.push_back(tol);
    this->move_to_angle_goal.goal_tolerance.push_back(tol);
  }

  trajectory_msgs::JointTrajectoryPoint point;
  point.time_from_start = ros::Duration(this->config.move_to_angles_time_from_start);
  this->move_to_angle_goal.goal_time_tolerance = point.time_from_start*this->config.move_to_angles_time_tol;
  // ROS_INFO_STREAM("Max time " << point.time_from_start);
  for (uint8_t i = 0; i < angles.size(); i++)
  {
    point.positions.push_back(angles[i]);
    point.velocities.push_back(vel[i]);
    //Just for tracking
    // ROS_INFO_STREAM("  joint[" << (int)i << "] error -> " << error[i] << "; vel -> " << error[i]/duration << "; time -> " << );
  }
  std::vector<trajectory_msgs::JointTrajectoryPoint>().swap(this->move_to_angle_goal.trajectory.points);
  this->move_to_angle_goal.trajectory.points.push_back(point);
  this->move_to_angle_goal.trajectory.header.stamp = ros::Time::now();
  this->new_angles = true;
  this->unlock();
  return true;
}

void CIriJointsModule::cancel_move_to_angles(void)
{
  if (!move_to_angle_action.is_enabled())
    return;
  if(this->state == IRI_JOINTS_MODULE_MOVE_ANGLES)
    this->cancel_move_to_angles_pending = true;
  else
    ROS_WARN("CIriJointsModule::cancel_move_to_angles-> MoveToAngles can't be cancelled because is not active.");
}

bool CIriJointsModule::is_move_to_angles_finished(void)
{
  if (!move_to_angle_action.is_enabled())
    return true;
  this->lock();
  if(this->state == IRI_JOINTS_MODULE_MOVE_ANGLES || (this->state == IRI_JOINTS_MODULE_IDLE && this->new_angles))
  {
    this->unlock();
    return false;
  }
  else
  {
    this->unlock();
    return true;
  }
}

/*Common*/
bool CIriJointsModule::set_tolerance(double tolerance)
{
  this->lock();
  if(!this->point_to_reconf.set_parameter("max_ik_tol",tolerance))
  {
    if(this->point_to_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status = IRI_JOINTS_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist
    {
      this->status = IRI_JOINTS_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  this->unlock();
  return true;
}

bool CIriJointsModule::set_base_frame(std::string &base_frame)
{
  this->lock();
  if(!this->point_to_reconf.set_parameter("base_frame",base_frame))
  {
    if(this->point_to_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status = IRI_JOINTS_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist
    {
      this->status = IRI_JOINTS_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  this->unlock();
  return true;
}

bool CIriJointsModule::set_urdf_param(std::string &urdf_param)
{
  this->lock();
  if(!this->point_to_reconf.set_parameter("urdf_param",urdf_param))
  {
    if(this->point_to_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status = IRI_JOINTS_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist
    {
      this->status = IRI_JOINTS_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  this->unlock();
  return true;
}

bool CIriJointsModule::set_parameter(const std::string &name,bool value)
{
  return false;
}

bool CIriJointsModule::set_parameter(const std::string &name,int value)
{
  this->lock();
  if(!this->point_to_reconf.set_parameter(name.c_str(), value))
  {
    if(this->point_to_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status = IRI_JOINTS_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist
    {
      this->status = IRI_JOINTS_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  this->unlock();
  return true;
}

bool CIriJointsModule::set_parameter(const std::string &name,std::string &value)
{
  this->lock();
  if(!this->point_to_reconf.set_parameter(name.c_str(), value))
  {
    if(this->point_to_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status = IRI_JOINTS_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist
    {
      this->status = IRI_JOINTS_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  this->unlock();
  return true;
}

bool CIriJointsModule::set_parameter(const std::string &name,double value)
{
  this->lock();
  if(!this->point_to_reconf.set_parameter(name.c_str(), value))
  {
    if(this->point_to_reconf.get_status()==DYN_RECONF_NO_CHANGE)
    {
      this->status = IRI_JOINTS_MODULE_SET_PARAM_FAIL;
      this->unlock();
      return false;
    }
    else// the parameter does not exist
    {
      this->status = IRI_JOINTS_MODULE_PARAM_NOT_PRESENT;
      this->unlock();
      return false;
    }
  }
  this->unlock();
  return true;
}

CIriJointsModule::~CIriJointsModule()
{

}
